import unittest

from bgez.core import CallbackList

from bgez.asyn import UNRESOLVED, Promise, Deferred
from bgez.asyn import LoopTestCase
from bgez import loop

class TestException(Exception):
    ...

class PromiseTest(LoopTestCase):

    def test_promise(self):
        reject_with = TestException('rejected')
        resolve_to = 3

        @Promise
        def will_resolve(resolve, reject):
            resolve(resolve_to)
        self._promise_resolve(will_resolve, resolve_to)

        @Promise
        def will_reject(resolve, reject):
            reject(reject_with)
        self._promise_reject(will_reject, reject_with)

    def test_promise_utility(self):
        reject_with = TestException('rejected')
        resolve_to = 3

        will_resolve = Promise.resolved(resolve_to)
        self._promise_resolve(will_resolve, resolve_to)

        will_reject = Promise.rejected(reject_with)
        self._promise_reject(will_reject, reject_with)

    def test_promise_peeking(self):
        deferred = Deferred()
        promise = deferred.promise

        self.assertEqual(UNRESOLVED, promise.peek())

        deferred.resolve()
        self.assertEqual(None, promise.peek())

        self._promise_resolve(promise, None)

    def test_promise_peeking_reject(self):
        reject_with = TestException('rejected')

        rejected = Promise.rejected(reject_with)
        self.assertRaises(TestException, rejected.peek)

        self._promise_reject(rejected, reject_with)

    def test_promise_resolve_chaining(self):
        a = object()
        b = object()

        self.assertNotEqual(a, b)

        promise = Promise.resolved(a) \
            .then(lambda v: b)

        @self.postLoop
        def _after_loop():
            self.assertEqual(b, promise.peek())

    def test_promise_catch_chaining(self):
        reject_with = TestException('chaining')
        resolve_to = object()

        promise = Promise.rejected(reject_with) \
            .catch(lambda e: resolve_to)

        @self.postLoop
        def _after_loop():
            self.assertEqual(resolve_to, promise.peek())

    def test_promise_wrapping(self):
        resolve_to = object()

        @Promise
        def a(resolve, _):
            resolve(resolve_to)

        @Promise
        def b(resolve, _):
            resolve(a)

        @Promise
        def c(resolve, _):
            resolve(b)

        self._promise_resolve(c, resolve_to)

    def test_promise_wrapping_reject(self):
        reject_with = TestException('meh')

        a = Promise.rejected(reject_with)
        b = Promise.resolved(a)
        c = Promise.resolved(b)

        self._promise_reject(c, reject_with)

    def test_promise_all(self):
        reject_with = TestException('meh')
        ok = object()

        a = Promise.resolved(1)
        b = Promise.resolved(2)
        c = Promise.resolved(3)
        d = Promise.rejected(reject_with)

        test1 = Promise.all([a, b, c])
        test2 = Promise.all([a, b, c, d]) \
            .catch(lambda exc: ok)

        @self.postLoop
        def _after_loop():
            self.assertListEqual([1, 2, 3], test1.peek())
            self.assertEqual(ok, test2.peek())

    def test_promise_race(self):
        reject_with = TestException('meh')
        resolve_to = object()
        ok = object()

        a = Promise.resolved(resolve_to)
        b = Promise(lambda a, b: None)
        c = Promise.rejected(reject_with)

        test1 = Promise.race([b, a])
        test2 = Promise.race([a, c])
        test3 = Promise.race([c, a]) \
            .catch(lambda e: ok)

        @self.postLoop
        def _after_loop():
            self.assertEqual(resolve_to, test1.peek())
            self.assertEqual(resolve_to, test2.peek())
            self.assertEqual(ok, test3.peek())

    def _promise_resolve(self, promise, expected):
        '''Helper function to test if a promise correctly resolved'''
        called = False

        def callback(resolved):
            nonlocal called
            self.assertEqual(resolved, expected)
            called = True

        promise.then(callback)
        promise.catch(lambda E, e: self.fail())
        self.assertFalse(called)

        @self.postLoop
        def was_it_called():
            self.assertTrue(called)

    def _promise_reject(self, promise, expected):
        '''Helper function to test if a promise correctly rejected'''
        called = False

        def callback(caught):
            nonlocal called
            self.assertEqual(caught, expected)
            called = True

        promise.then(lambda v: self.fail())
        promise.catch(callback)
        self.assertFalse(called)

        @self.postLoop
        def was_it_called():
            self.assertTrue(called)
