from typing import Callable, Iterable, Coroutine

import time
import math

from collections import OrderedDict, deque

from bgez.core.abstract import Abstract, abstractmethod
from bgez.asyn import BasePromise, Executor, iscoroutine

__all__ = [
    'BaseLoop', 'EventLoop',
]

class BaseLoop(Abstract):
    '''BaseLoop class to inherit from, every method must be implemented'''

    @abstractmethod
    def register(self, coroutine: Coroutine) -> BasePromise:
        '''Wraps a coroutine-like object in a promise.'''

    @abstractmethod
    def immediate(self, callback: Callable, *args: list, **kargs: dict) -> None:
        '''Executes `callback(*args, **kargs)` as soon as possible.'''

    @abstractmethod
    def callback(self, callback: Callable, *args: list, **kargs: dict) -> None:
        '''Executes a callback on the callback step.'''

    @abstractmethod
    def step(self) -> bool:
        '''
        Run one cycle of the event loop.
        Returns `True` if everything was processed, else `False`.
        '''

    @abstractmethod
    def run(self, timeout: float = None, clock = time.monotonic, deadline: float = math.inf) -> bool:
        '''
        Run the event loop continuously, until `deadline` is passed,
        or until no more events are to be processed.\n
        Returns `True` if everything was processed, else `False`.
        '''

class EventLoop(BaseLoop):

    def __init__(self, processors: Iterable[Callable]):
        self.processors = processors
        self._immediates = deque()
        self._callbacks = deque()

    def _run_callbacks(self, callbacks: deque):
        while callbacks:
            callback, args, kargs = callbacks.popleft()
            callback(*args, **kargs)

    def register(self, coroutine: Coroutine) -> BasePromise:
        if not iscoroutine(coroutine):
            raise ValueError(f'can only register a coroutine, got {type(object)}')
        executor = Executor(coroutine)
        executor.send(None)
        return executor.promise

    def immediate(self, callback: Callable, *args: list, **kargs: dict):
        self._immediates.append((callback, args, kargs))

    def callback(self, callback: Callable, *args: list, **kargs: dict):
        self._callbacks.append((callback, args, kargs))

    def step(self) -> bool:
        empty = True

        for processor in self.processors:
            self._run_callbacks(self._immediates)
            if not processor() and empty:
                empty = False

        self._run_callbacks(self._immediates)
        self._run_callbacks(self._callbacks)
        self._run_callbacks(self._immediates)

        return empty

    def run(self, timeout: float = None, clock = time.monotonic, deadline: float = math.inf) -> bool:
        if timeout is not None:
            deadline = timeout + clock()
        while deadline > clock():
            if self.step(): return True
        return False
