from typing import Any, Coroutine

from bgez.core.abstract import Abstract, abstractmethod
from bgez.asyn import BasePromise, Promise, Deferred

__all__ = [
    'BaseExecutor', 'Executor',
]

class BaseExecutor(Abstract):
    '''
    Wraps a coroutine-like object and drives its execution.
    '''

    @property
    @abstractmethod
    def promise(self) -> BasePromise:
        '''Get the associated promise for this executor.'''

    @abstractmethod
    def send(self, value: Any = None) -> None:
        '''Send a value to the driven coroutine.'''

    @abstractmethod
    def throw(self, exc: Exception) -> None:
        '''Throw an error to the driven coroutine.'''

class Executor(BaseExecutor):

    __slots__ = [
        '__deferred',
        '__coroutine',
    ]

    def __init__(self, coroutine: Coroutine):
        self.__deferred = Deferred()
        self.__coroutine = coroutine

    @property
    def promise(self):
        return self.__deferred.promise

    def send(self, value: Any) -> None:
        try:
            callback = self.__coroutine.send(value)
            callback(self)
        except StopIteration as stop:
            self.__deferred.resolve(stop.value)
        except Exception as e:
            self.__deferred.reject(e)

    def throw(self, exc: Exception) -> None:
        try:
            callback = self.__coroutine.throw(exc)
            callback(self)
        except StopIteration as stop:
            self.__deferred.resolve(stop.value)
        except Exception as e:
            self.__deferred.reject(e)
