from typing import Any, List, Callable, Iterable

from collections import defaultdict

from bgez.core.abstract import Abstract, abstractmethod

__all__ = [
    'InputType',
    'Input',
    'InputAdapter',
]

class InputType(Abstract):
    '''
    A value type represents a specific input space.
    e.g.: A joystick returns a float when a button returns an int.
    '''

    @staticmethod
    def filter(values) -> Iterable[Any]:
        '''Removes `None` values from a sequence.'''
        return filter(lambda value: value is not None, values)

    @classmethod
    @abstractmethod
    def union(cls, values: List[Any]) -> Any:
        '''Returns the union for the inputs'''

    @classmethod
    @abstractmethod
    def intersection(cls, values: List[Any]) -> Any:
        '''Returns the intersection for the inputs'''

    @classmethod
    def invoke(cls, method: str, value: Any) -> Any:
        '''Tries to call the requested `method` with `value`.'''
        return getattr(cls, method)(value)

class Input(Abstract):
    '''
    Inputs are device-bound objects that return the state for some key.
    Inputs return only one kind of InputType.

    Defines two classmethods: `union` and `intersection`.
    '''

    def __getattr__(self, method):
        '''Proxy access to some method to the current InputType.'''
        return self.type.invoke(method, self.value)

    @property
    def type(self) -> InputType:
        try:
            return getattr(self.__class__, 'TYPE')
        except LookupError:
            ...
        raise TypeError('Input derived classes should define a `TYPE` member')

    @property
    @abstractmethod
    def value(self) -> Any:
        '''Returns the value for this input, `None` should be discarded.'''

class InputAdapter(Input):
    '''Wraps some input to convert its values to some other InputType value.'''

    def __init__(self, input: Input, target: InputType, converter: Callable):
        self.converter = converter
        self.target = target
        self.input = input

    @property
    def type(self):
        return self.target

    @property
    def value(self):
        return self.converter(self.input.value)
