from typing import Any, Callable, Hashable, Iterable, Dict, Tuple, List

import json

from collections import OrderedDict, defaultdict
from itertools import chain

from bgez.core.abstract import Abstract, abstractmethod
from bgez.core import Dictable, Saveable, decorator
from bgez.core.input import InputType, Input
from bgez import logger

__all__ = [
    'BaseKeymapEntry', 'KeymapEntry',
    'BaseKeymap', 'Keymap',
]

def input_value(input: Input) -> Any:
    return input.value

class BaseKeymapEntry(Input):
    '''
    KeymapEntries are the real process unit for inputs.
    They manage the list of bindings assigned to an action, and perform the
    conversion / merging of the values, based on the input type.
    '''

    @abstractmethod
    def __init__(self, name: str, keymap: 'BaseKeymap', type: InputType, converters: Dict[type, Callable]):
        '''KeymapEntries depend on an input registry and an InputType.'''

    @abstractmethod
    def __len__(self):
        '''Number of bindings for this entry.'''

    @abstractmethod
    def __contains__(self, binding: str):
        '''Returns if this entry contains `binding`.'''

    @abstractmethod
    def link(self, action: Input, linker: Callable) -> None:
        '''Uses `linker` so that `action` contributes to this input.'''

    @property
    @abstractmethod
    def name(self):
        '''The name given by the keymap to this entry.'''

    @property
    @abstractmethod
    def type(self):
        '''The `InputType` instance for this entry.'''

    @property
    @abstractmethod
    def bindings(self) -> Iterable[str]:
        '''Returns the list of bindings for this entry.'''

    @property
    def links(self) -> Iterable[str]:
        '''Returns the list of linked actions to this entry.'''

    @abstractmethod
    def set(self, bindings: List[str]) -> None:
        '''Sets the bindings for this entry.'''

    @abstractmethod
    def add(self, *bindings: Tuple[str]) -> None:
        '''Adds `bindings` to this entry.'''

    @abstractmethod
    def remove(self, *bindings: Tuple[str]) -> None:
        '''Removes `bindings` from this entry.'''

class BaseKeymap(Dictable, Saveable, Abstract):
    '''
    KeyMaps make the link between `actions` and `keybindings`.
    This is the high level abstraction for input management.
    '''

    @abstractmethod
    def __init__(self, registry: Dict[str, Input], *, defaults: Dict[str, List[str]] = {}):
        '''Keymaps should take an input registry as input.'''

    @abstractmethod
    def __contains__(self, action: Hashable) -> bool:
        '''Returns if `action` is defined for this keymap.'''

    @property
    @abstractmethod
    def actions(self) -> Iterable[Hashable]:
        '''Returns the list of registered actions.'''

    @abstractmethod
    def input(self, name: str) -> Input:
        '''Returns the input from the registry for `name`.'''

    @abstractmethod
    def define(self,
        action: Hashable, type: InputType, *,
        converters: Dict[type, Callable] = {}, default: List[str] = []
    ) -> BaseKeymapEntry:
        '''
        Defines an action and the type of expected value.\n
        - `action`: name of the action to define.
        - `type`: the type of expected values.
        - `converters`: the map of incompatible types -> conversion function.
        - `default`: the default bindings if not defined in the config.
        '''

    @abstractmethod
    def link(self, action: Hashable, sub_action: Hashable, linker: Callable = input_value):
        '''Creates a link between two actions.'''

    @abstractmethod
    def get(self, action: Hashable) -> BaseKeymapEntry:
        '''Returns the list of keybindings for the action.'''

    @abstractmethod
    def set(self, action: Hashable, bindings: List[str]):
        '''Sets the bindings for `action`.'''

    @abstractmethod
    def add(self, action: Hashable, *keybindings: Tuple[str]) -> Input:
        '''Adds a keybinding for the action.'''

    @abstractmethod
    def remove(self, action: Hashable, keybinding: str) -> None:
        '''Removes a keybinding from an action.'''

    @abstractmethod
    def typeof(self, action: Hashable) -> InputType:
        '''Returns the associated type for `action`.'''

    @abstractmethod
    def valueof(self, action: Hashable) -> Any:
        '''Returns the value for a given action.'''

class KeymapEntry(BaseKeymapEntry):

    def __init__(self, name, keymap, type, converters = {}):
        self._converters = converters.copy()
        self._bindings = OrderedDict()
        self._links = OrderedDict()
        self._keymap = keymap
        self._name = name
        self._type = type

    def __len__(self):
        return len(self._bindings.keys())

    def __contains__(self, binding):
        binding, _ = self._parse(binding)
        return binding in self._bindings

    def _get_input(self, name: str) -> Input:
        return self._keymap.input(name)

    def _parse(self, keybinding: str) -> Tuple[str, List[str]]:
        keys = [key.strip() for key in keybinding.split('+')]
        keys.sort()
        return '+'.join(keys), keys

    def _converter(self, input_type: type) -> Callable:
        '''Gets the conversion function(`input_type`) => type(self)'''
        if input_type == self.type:
            return input_value
        try:
            return self._converters[input_type]
        except KeyError:
            ...
        raise TypeError(f'no converters found for {input_type} to {type(self)}')

    def _resolver(self, input: Input) -> Callable:
        '''Returns a function that will always return the good input type for `input.value`'''
        return (lambda input, converter:
            lambda: converter(input)
        )(input, self._converter(input.type))

    def link(self, action, linker = input_value):
        if isinstance(action, str):
            action = self._keymap.get(action)
        self._links[action.name] = lambda: linker(action)

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    @property
    def bindings(self):
        return tuple(self._bindings.keys())

    @property
    def links(self):
        return tuple(self._links.keys())

    def set(self, bindings):
        self._bindings.clear()
        self.add(*bindings)

    def add(self, *bindings):
        for binding in bindings:
            binding, keys = self._parse(binding)

            resolvers = [self._resolver(self._get_input(key))
                for key in keys]

            state = (lambda resolvers, intersection, filter: # hmmm
                lambda: intersection(filter(resolver() for resolver in resolvers))
            )(resolvers, self.type.intersection, self.type.filter)

            self._bindings[binding] = state

    def remove(self, *bindings):
        for binding in bindings:
            binding, _ = self._parse(binding)
            del self._bindings[binding]

    @property
    def value(self):
        type = self.type; return type.union(type.filter(
            input() for input in chain(self._links.values(), self._bindings.values())
        ))

class Keymap(BaseKeymap):
    '''Default implementation for BaseKeymap.'''

    Entry = KeymapEntry

    def __init__(self, registry: Dict[str, Input], *, defaults: Dict[str, List[str]] = {}):
        self._defaults = defaults.copy()
        self._mappings = OrderedDict()
        self._registry = registry

    def __contains__(self, action):
        return action in self._mappings

    def __getitem__(self, action):
        return self.get(action)

    def _resolve_path(self, path: str) -> str:
        return path

    @property
    def actions(self):
        return tuple(self._mappings.keys())

    def input(self, name):
        return self._registry[name]

    def define(self, action, type, *, converters = {}, default = []):
        if action in self: raise ValueError(f'"{action}" is already defined')
        self._mappings[action] = entry = self.Entry(action, self, type, converters)
        bindings = self._defaults[action] = default.copy()
        entry.set(bindings)
        return entry

    @decorator
    def link(self, action, sub_action, linker = None):
        if isinstance(sub_action, str):
            sub_action = self.get(sub_action)
        if isinstance(action, str):
            action = self.get(action)
        action.link(sub_action, linker)

    def get(self, action):
        return self._mappings[action]

    def set(self, action, keybindings):
        self.get(action).set(keybindings)

    def add(self, action, *keybindings):
        self.get(action).add(*keybindings)

    def remove(self, action, *keybindings):
        self.get(action).remove(*keybindings)

    def typeof(self, action):
        return self.get(action).type

    def valueof(self, action):
        return self.get(action).value

    def to_dict(self):
        defaulted = self._defaults.copy()
        defaulted.update({
            action: entry.bindings
            for action, entry in self._mappings.items()
        })
        return defaulted

    def from_dict(self, data):
        defaulted = self._defaults.copy()
        defaulted.update(data)

        for action, bindings in defaulted.items():
            if action in self:
                self.set(action, bindings)
            else:
                logger.warning('"%s" is not defined in %s', action, self)

    def save(self, path, *, opener = open):
        resolved_path = self._resolve_path(path)
        with opener(resolved_path, 'w') as file:
            json.dump({
                'actions': self.to_dict(),
                # 'metadata': {},
            }, file, indent=2)

    def load(self, path, *, opener = open):
        resolved_path = self._resolve_path(path)
        try:
            with opener(resolved_path, 'r') as file:
                data = json.load(file)
        except FileNotFoundError:
            logger.info('file "%s" doesn\'t exist, using default keymap bindings', path)
            data = {}

        self.from_dict(data.get('actions', {}))
        self.save(path, opener=opener)
