from typing import Callable

from weakref import finalize

from bgez.core.events import EventEmitter
from bgez.asyn import BaseLoop, Promise
from bgez.core.time import Scheduler
from bgez import loop

from bgez.logic import SceneManager

from bge import logic # pylint: disable=E0401

__all__ = [
    'Game',
]

class Game:
    '''Game API'''

    def __init__(self, *, time: Scheduler, scenes: SceneManager):
        self.__events = EventEmitter(['start', 'error', 'quit', 'exit'])
        self.__exit = finalize(logic, self.__events.emit, 'exit')
        self.__handlers = [lambda: loop.get().step()]
        self.__resources = []
        self.__time = time

        self.scenes = scenes

    def _emit_start(self):
        self.__events.emit('start', self.scenes.current)

    def _throw_error(self, e):
        self.__events.emit('error', e)

    def _step(self) -> None:
        '''Executes one logic step.'''
        for handler in self.__handlers:
            handler()

    def on(self, *args):
        '''Registers your callbacks to a game event.'''
        return self.__events.once(*args)

    def use(self, handler: Callable) -> None:
        '''Use a new handler triggered on each step.'''
        self.__handlers.append(handler)

    def timeout(self, timeout: float, callback: Callable, *args, **kargs) -> None:
        '''Register a callback to be executed later in time, `timeout` in seconds.'''
        self.__time.timeout(timeout, callback, *args, **kargs)

    def interval(self, interval: float, callback: Callable, *args, **kargs) -> None:
        '''Register a callback to be executed periodicaly over time, `interval` in seconds.'''
        self.__time.interval(interval, callback, *args, **kargs)

    def sleep(self, time: float):
        return Promise(lambda resolve, _: self.timeout(time, resolve))

    def skip(self, frames: int):
        return Promise(lambda resolve, _: self.scenes.current.logic.timeout(frames, resolve))

    def restart(self) -> None:
        '''Restarts the current game.'''
        logic.restartGame()

    def start(self, file: str) -> None:
        '''Starts a different `.blend` file.'''
        logic.startGame(file)

    def quit(self) -> None:
        '''Quits the game.'''
        self.__events.emit('quit')
        logic.endGame()

    def load(self, path: str, kind: str = 'Scene') -> Promise:
        '''Loads a resource at `path`.'''
        def resolver(resolve, _):
            status = logic.LibLoad(path, kind, asynchronous=True)
            status.onFinish = lambda status: resolve(status.timeTaken)
        return Promise(resolver)

    def unload(self, resource) -> None:
        '''Unloads `resource`.'''

    @property
    def resources(self) -> list:
        '''List of the loaded resources.'''
        return self.__resources

    @property
    def gravity(self) -> float:
        '''The game gravity setting.'''
        return logic.getGravity()

    @gravity.setter
    def gravity(self, value: float):
        logic.setGravity(value)
