import math

from collections import namedtuple

from bgez.core.utils import Singleton

from bge import logic, render # pylint: disable=E0401

__all__ = [
    'Window',
]

_screen = namedtuple('Screen', 'width height')
_position = namedtuple('Position', 'x y')

class Window:

    __debug = dict(
        framerate = render.showFramerate,
        profile = render.showProfile,
        properties = render.showProperties,
        autoDebugList = render.autoDebugList,
    )

    def getDisplay(self, index: int = 0):
        return _screen(*render.getDisplayDimensions())

    def setSize(self, width: int, height: int) -> _position:
        render.setWindowSize(width, height)
        return width, height

    @property
    def displayCount(self):
        return 1

    @property
    def displays(self):
        return tuple(self.getDisplay(i) for i in range(self.displayCount))

    @property
    def size(self):
        return self.width, self.height

    @size.setter
    def size(self, value):
        self.setSize(*value)

    @property
    def width(self) -> int:
        return render.getWindowWidth()

    @width.setter
    def width(self, value: int) -> None:
        self.setSize(value, self.height)

    @property
    def height(self) -> int:
        return render.getWindowHeight()

    @height.setter
    def height(self, value: int) -> None:
        self.setSize(self.width, value)

    @property
    def fullscreen(self) -> bool:
        return render.getFullScreen()

    @fullscreen.setter
    def fullscreen(self, value: bool) -> None:
        render.setFullscreen(value)

    @property
    def anisotropic(self) -> bool:
        render.getAnisotropicFilter()

    @anisotropic.setter
    def anisotropic(self, value: bool) -> None:
        render.setAnisotropicFilter(value)

    def getNormalizedPosition(self, x: float, y: float) -> (float, float):
        '''
        Returns the normalized relative position in the window for `x` and `y`.
        '''
        w, h = self.width - 1, self.height - 1
        return _position(int(x * w) / w, int(y * h) / h)

    def getMousePosition(self, format: str = '%') -> _position:
        '''
        Returns the mouse position given a format:

        >>> getMousePosition('%')
        (0.2, 1.0)
        >>> getMousePosition('px')
        (134, 2)
        '''
        position = _position(logic.mouse.position)
        if format == '%':
            return position
        elif format == 'px':
            return _position(
                x = position.x * (self.width - 1),
                y = position.y * (self.height - 1)
            )
        raise ValueError(f'"{format}" should be one of (%, px)')

    @property
    def mouse(self) -> _position:
        return _position(*logic.mouse.position)

    @mouse.setter
    def mouse(self, value: (float, float)) -> None:
        x, y = value
        if isinstance(x, int): x /= self.width
        if isinstance(y, int): y /= self.height
        logic.mouse.position = x, y

    def screenshot(self, name: str):
        path = logic.expandPath(name)
        render.makeScreenshot(path)
        return path

    def show(self, **kargs) -> None:
        for key, value in kargs.items():
            self.__debug[key](value)
